@@ -32,10 +32,12 @@ struct l_cipher;
 enum l_cipher_type {
 	L_CIPHER_AES = 0,
 	L_CIPHER_AES_CBC,
+	L_CIPHER_AES_CTR,
 	L_CIPHER_ARC4,
 	L_CIPHER_DES,
 	L_CIPHER_DES_CBC,
 	L_CIPHER_DES3_EDE_CBC,
+	L_CIPHER_RC2_CBC,
 };
 
 struct l_cipher *l_cipher_new(enum l_cipher_type type,
@@ -62,6 +64,7 @@ struct l_aead_cipher;
 
 enum l_aead_cipher_type {
 	L_AEAD_CIPHER_AES_CCM = 0,
+	L_AEAD_CIPHER_AES_GCM,
 };
 
 struct l_aead_cipher *l_aead_cipher_new(enum l_aead_cipher_type type,
