@@ -24,6 +24,7 @@
 #include <config.h>
 #endif
 
+#include <linux/types.h>
 #include <net/ethernet.h>
 #include <sys/socket.h>
 #include <arpa/inet.h>
@@ -38,6 +39,7 @@
 #include "util.h"
 #include "strv.h"
 #include "timeout.h"
+#include "acd.h"
 
 /* 8 hours */
 #define DEFAULT_DHCP_LEASE_SEC (8*60*60)
@@ -79,6 +81,8 @@ struct l_dhcp_server {
 	l_dhcp_destroy_cb_t event_destroy;
 
 	struct dhcp_transport *transport;
+
+	struct l_acd *acd;
 };
 
 #define MAC "%02x:%02x:%02x:%02x:%02x:%02x"
@@ -806,6 +810,21 @@ LIB_EXPORT bool l_dhcp_server_start(struct l_dhcp_server *server)
 
 	server->started = true;
 
+	server->acd = l_acd_new(server->ifindex);
+	l_acd_set_skip_probes(server->acd, true);
+	l_acd_set_defend_policy(server->acd, L_ACD_DEFEND_POLICY_INFINITE);
+
+	ia.s_addr = server->address;
+
+	/* In case of unit testing we don't want this to be a fatal error */
+	if (!l_acd_start(server->acd, inet_ntoa(ia))) {
+		SERVER_DEBUG("Failed to start ACD on %s, continuing without",
+				IP_STR(server->address));
+
+		l_acd_destroy(server->acd);
+		server->acd = NULL;
+	}
+
 	return true;
 }
 
@@ -827,6 +846,11 @@ LIB_EXPORT bool l_dhcp_server_stop(struct l_dhcp_server *server)
 		server->next_expire = NULL;
 	}
 
+	if (server->acd) {
+		l_acd_destroy(server->acd);
+		server->acd = NULL;
+	}
+
 	/* TODO: Add ability to save leases */
 
 	return true;
