@@ -29,57 +29,31 @@
 #include <linux/types.h>
 #include <net/if_arp.h>
 #include <errno.h>
-#include <time.h>
+#include <arpa/inet.h>
 
 #include "private.h"
 #include "random.h"
+#include "time.h"
+#include "time-private.h"
 #include "net.h"
 #include "timeout.h"
 #include "dhcp.h"
 #include "dhcp-private.h"
+#include "netlink.h"
+#include "rtnl.h"
+#include "acd.h"
+
+#define CLIENT_DEBUG(fmt, args...)					\
+	l_util_debug(client->debug_handler, client->debug_data,		\
+			"%s:%i " fmt, __func__, __LINE__, ## args)
+#define CLIENT_ENTER_STATE(s)						\
+	l_util_debug(client->debug_handler, client->debug_data,		\
+			"%s:%i Entering state: " #s,			\
+			__func__, __LINE__);				\
+	client->state = (s)
 
-#define DHCP_MAGIC 0x63825363
 #define BITS_PER_LONG (sizeof(unsigned long) * 8)
 
-#define DHCP_OPTION_PAD 0 /* RFC 2132, Section 3.1 */
-#define DHCP_OPTION_END 255 /* RFC 2132, Section 3.2 */
-
-/* RFC 2132, Section 9.3. Option Overload */
-#define DHCP_OPTION_OVERLOAD 52
-enum dhcp_option_overload {
-	DHCP_OVERLOAD_FILE = 1,
-	DHCP_OVERLOAD_SNAME = 2,
-	DHCP_OVERLOAD_BOTH = 3,
-};
-
-/* RFC 2132, Section 9.6. DHCP Message Type */
-#define DHCP_OPTION_MESSAGE_TYPE 53
-enum dhcp_message_type {
-	DHCP_MESSAGE_TYPE_DISCOVER = 1,
-	DHCP_MESSAGE_TYPE_OFFER = 2,
-	DHCP_MESSAGE_TYPE_REQUEST = 3,
-	DHCP_MESSAGE_TYPE_DECLINE = 4,
-	DHCP_MESSAGE_TYPE_ACK = 5,
-	DHCP_MESSAGE_TYPE_NAK = 6,
-	DHCP_MESSAGE_TYPE_RELEASE = 7,
-	DHCP_MESSAGE_TYPE_INFORM = 8,
-};
-
-#define DHCP_OPTION_PARAMETER_REQUEST_LIST 55 /* Section 9.8 */
-#define DHCP_OPTION_MAXIMUM_MESSAGE_SIZE 57 /* Section 9.10 */
-#define DHCP_OPTION_CLIENT_IDENTIFIER 61 /* Section 9.14 */
-
-/* RFC 2131, Table 1 */
-enum dhcp_op_code {
-	DHCP_OP_CODE_BOOTREQUEST = 1,
-	DHCP_OP_CODE_BOOTREPLY = 2,
-};
-
-enum {
-	DHCP_PORT_SERVER = 67,
-	DHCP_PORT_CLIENT = 68,
-};
-
 enum dhcp_state {
 	DHCP_STATE_INIT,
 	DHCP_STATE_SELECTING,
@@ -156,229 +130,7 @@ const char *_dhcp_option_to_string(uint8_t option)
 		return "End";
 	default:
 		return "unknown";
-	};
-}
-
-bool _dhcp_message_iter_init(struct dhcp_message_iter *iter,
-				const struct dhcp_message *message, size_t len)
-{
-	if (!message)
-		return false;
-
-	if (len < sizeof(struct dhcp_message) + 4)
-		return false;
-
-	if (l_get_be32(message->options) != DHCP_MAGIC)
-		return false;
-
-	memset(iter, 0, sizeof(*iter));
-	iter->message = message;
-	iter->message_len = len;
-	iter->pos = 4;
-	iter->max = len - sizeof(struct dhcp_message);
-	iter->options = message->options;
-	iter->can_overload = true;
-
-	return true;
-}
-
-static bool next_option(struct dhcp_message_iter *iter,
-				uint8_t *t, uint8_t *l, const void **v)
-{
-	uint8_t type;
-	uint8_t len;
-
-	while (iter->pos < iter->max) {
-		type = iter->options[iter->pos];
-
-		switch (type) {
-		case DHCP_OPTION_PAD:
-			iter->pos += 1;
-			continue;
-		case DHCP_OPTION_END:
-			return false;
-		default:
-			break;
-		}
-
-		if (iter->pos + 2 >= iter->max)
-			return false;
-
-		len = iter->options[iter->pos + 1];
-
-		if (iter->pos + 2 + len > iter->max)
-			return false;
-
-		*t = type;
-		*l = len;
-		*v = &iter->options[iter->pos + 2];
-
-		iter->pos += 2 + len;
-		return true;
 	}
-
-	return false;
-}
-
-bool _dhcp_message_iter_next(struct dhcp_message_iter *iter, uint8_t *type,
-				uint8_t *len, const void **data)
-{
-	bool r;
-	uint8_t t, l;
-	const void *v;
-
-	do {
-		r = next_option(iter, &t, &l, &v);
-		if (!r) {
-			iter->can_overload = false;
-
-			if (iter->overload_file) {
-				iter->options = iter->message->file;
-				iter->pos = 0;
-				iter->max = sizeof(iter->message->file);
-				iter->overload_file = false;
-				continue;
-			}
-
-			if (iter->overload_sname) {
-				iter->options = iter->message->sname;
-				iter->pos = 0;
-				iter->max = sizeof(iter->message->sname);
-				iter->overload_sname = false;
-				continue;
-			}
-
-			return r;
-		}
-
-		switch (t) {
-		case DHCP_OPTION_OVERLOAD:
-			if (l != 1)
-				continue;
-
-			if (!iter->can_overload)
-				continue;
-
-			if (l_get_u8(v) & DHCP_OVERLOAD_FILE)
-				iter->overload_file = true;
-
-			if (l_get_u8(v) & DHCP_OVERLOAD_SNAME)
-				iter->overload_sname = true;
-
-			continue;
-		default:
-			if (type)
-				*type = t;
-
-			if (len)
-				*len = l;
-
-			if (data)
-				*data = v;
-			return r;
-		}
-	} while (true);
-
-	return false;
-}
-
-int _dhcp_option_append(uint8_t **buf, size_t *buflen, uint8_t code,
-					size_t optlen, const void *optval)
-{
-	if (!buf || !buflen)
-		return -EINVAL;
-
-	switch (code) {
-
-	case DHCP_OPTION_PAD:
-	case DHCP_OPTION_END:
-		if (*buflen < 1)
-			return -ENOBUFS;
-
-		(*buf)[0] = code;
-		*buf += 1;
-		*buflen -= 1;
-		break;
-
-	default:
-		if (*buflen < optlen + 2)
-			return -ENOBUFS;
-
-		if (!optval)
-			return -EINVAL;
-
-		(*buf)[0] = code;
-		(*buf)[1] = optlen;
-		memcpy(&(*buf)[2], optval, optlen);
-
-		*buf += optlen + 2;
-		*buflen -= (optlen + 2);
-
-		break;
-	}
-
-	return 0;
-}
-
-static int dhcp_append_prl(const unsigned long *reqopts,
-					uint8_t **buf, size_t *buflen)
-{
-	uint8_t optlen = 0;
-	unsigned int i;
-	unsigned int j;
-
-	if (!buf || !buflen)
-		return -EINVAL;
-
-	for (i = 0; i < 256 / BITS_PER_LONG; i++)
-		optlen += __builtin_popcountl(reqopts[i]);
-
-	/*
-	 * This function assumes that there's enough space to put the PRL
-	 * into the buffer without resorting to file or sname overloading
-	 */
-	if (*buflen < optlen + 2U)
-		return -ENOBUFS;
-
-	i = 0;
-	(*buf)[i++] = DHCP_OPTION_PARAMETER_REQUEST_LIST;
-	(*buf)[i++] = optlen;
-
-	for (j = 0; j < 256; j++) {
-		if (reqopts[j / BITS_PER_LONG] & 1UL << (j % BITS_PER_LONG))
-			(*buf)[i++] = j;
-	}
-
-	*buf += optlen + 2;
-	*buflen -= (optlen + 2);
-
-	return 0;
-}
-
-static int dhcp_message_init(struct dhcp_message *message,
-				enum dhcp_op_code op,
-				uint8_t type, uint32_t xid,
-				uint8_t **opt, size_t *optlen)
-{
-	int err;
-
-	message->op = op;
-	message->xid = L_CPU_TO_BE32(xid);
-
-	if (*optlen < 4)
-		return -ENOBUFS;
-
-	*optlen -= 4;
-	*opt = (uint8_t *)(message + 1);
-	l_put_be32(DHCP_MAGIC, *opt);
-	*opt += 4;
-
-	err = _dhcp_option_append(opt, optlen,
-					DHCP_OPTION_MESSAGE_TYPE, 1, &type);
-	if (err < 0)
-		return err;
-
-	return 0;
 }
 
 static void dhcp_message_set_address_type(struct dhcp_message *message,
@@ -415,8 +167,6 @@ static inline int dhcp_message_optimize(struct dhcp_message *message,
 	return len;
 }
 
-#define DHCP_MIN_OPTIONS_SIZE 312
-
 struct l_dhcp_client {
 	enum dhcp_state state;
 	unsigned long request_options[256 / BITS_PER_LONG];
@@ -428,12 +178,21 @@ struct l_dhcp_client {
 	char *hostname;
 	uint32_t xid;
 	struct dhcp_transport *transport;
-	time_t start_t;
+	uint64_t start_t;
 	struct l_timeout *timeout_resend;
+	struct l_timeout *timeout_lease;
 	struct l_dhcp_lease *lease;
+	struct l_netlink *rtnl;
+	uint32_t rtnl_add_cmdid;
+	struct l_rtnl_address *rtnl_configured_address;
+	uint8_t attempt;
 	l_dhcp_client_event_cb_t event_handler;
 	void *event_data;
 	l_dhcp_destroy_cb_t event_destroy;
+	l_dhcp_debug_cb_t debug_handler;
+	l_dhcp_destroy_cb_t debug_destroy;
+	struct l_acd *acd;
+	void *debug_data;
 	bool have_addr : 1;
 	bool override_xid : 1;
 };
@@ -445,10 +204,10 @@ static inline void dhcp_enable_option(struct l_dhcp_client *client,
 						1UL << (option % BITS_PER_LONG);
 }
 
-static uint16_t dhcp_attempt_secs(time_t start)
+static uint16_t dhcp_attempt_secs(uint64_t start)
 {
-	time_t now = time(NULL);
-	time_t elapsed = now - start;
+	uint64_t now = l_time_now();
+	uint64_t elapsed = l_time_to_secs(now - start);
 
 	if (elapsed == 0)
 		return 1;
@@ -459,18 +218,78 @@ static uint16_t dhcp_attempt_secs(time_t start)
 	return elapsed;
 }
 
+/*
+ * Takes a time in seconds and produces a fuzzed value that can be directly
+ * used by l_timeout_modify_ms
+ */
+static uint64_t dhcp_fuzz_secs(uint32_t secs)
+{
+	/*
+	 * RFC2132, Section 4.1:
+	 * DHCP clients are responsible for all message retransmission.  The
+	 * client MUST adopt a retransmission strategy that incorporates a
+	 * randomized exponential backoff algorithm to determine the delay
+	 * between retransmissions.
+	 *
+	 * and later in the same paragraph:
+	 * For example, in a 10Mb/sec Ethernet internetwork, the delay before
+	 * the first retransmission SHOULD be 4 seconds randomized by the
+	 * value of a uniform random number chosen from the range -1 to +1.
+	 * Clients with clocks that provide resolution granularity of less than
+	 * one second may choose a non-integer randomization value.
+	 */
+	return _time_fuzz_secs(secs, 1);
+}
+
+/*
+ * Takes a time in milliseconds and produces a fuzzed value that can be directly
+ * used by l_timeout_modify_ms. The fluctuation of the random noise added is
+ * from -63 to 63 milliseconds.
+ */
+static uint64_t dhcp_fuzz_msecs(uint64_t ms)
+{
+	uint32_t r = l_getrandom_uint32();
+
+	if (r & 0x80000000)
+		ms += r & 0x3f;
+	else
+		ms -= r & 0x3f;
+
+	return ms;
+}
+
+static uint32_t dhcp_rebind_renew_retry_time(uint64_t start_t, uint32_t expiry)
+{
+	uint64_t now = l_time_now();
+	uint32_t relative_now;
+	uint32_t retry_time;
+
+	/*
+	 * RFC 2131, Section 4.4.5:
+	 * "   In both RENEWING and REBINDING states, if the client receives no
+	 * response to its DHCPREQUEST message, the client SHOULD wait one-half
+	 * of the remaining time until T2 (in RENEWING state) and one-half of
+	 * the remaining lease time (in REBINDING state), down to a minimum of
+	 * 60 seconds, before retransmitting the DHCPREQUEST message.
+	 */
+	relative_now = l_time_to_secs(now - start_t);
+	retry_time = (expiry - relative_now) / 2;
+
+	if (retry_time < 60)
+		retry_time = 60;
+
+	return retry_time;
+}
+
 static int client_message_init(struct l_dhcp_client *client,
 					struct dhcp_message *message,
-					uint8_t type,
-					uint8_t **opt, size_t *optlen)
+					struct dhcp_message_builder *builder)
 {
-	int err;
 	uint16_t max_size;
 
-	err = dhcp_message_init(message, DHCP_OP_CODE_BOOTREQUEST,
-				type, client->xid, opt, optlen);
-	if (err < 0)
-		return err;
+	message->op = DHCP_OP_CODE_BOOTREQUEST;
+	message->xid = L_CPU_TO_BE32(client->xid);
+	message->magic = L_CPU_TO_BE32(DHCP_MAGIC);
 
 	dhcp_message_set_address_type(message, client->addr_type,
 							client->addr_len);
@@ -491,20 +310,19 @@ static int client_message_init(struct l_dhcp_client *client,
 	 */
 	message->secs = L_CPU_TO_BE16(dhcp_attempt_secs(client->start_t));
 
-	err = dhcp_append_prl(client->request_options, opt, optlen);
-	if (err < 0)
-		return err;
+	if (!_dhcp_message_builder_append_prl(builder,
+						client->request_options))
+		return -EINVAL;
 
 	/*
 	 * Set the maximum DHCP message size to the minimum legal value.  This
 	 * helps some buggy DHCP servers to not send bigger packets
 	 */
 	max_size = L_CPU_TO_BE16(576);
-	err = _dhcp_option_append(opt, optlen,
+	if (!_dhcp_message_builder_append(builder,
 					DHCP_OPTION_MAXIMUM_MESSAGE_SIZE,
-					2, &max_size);
-	if (err < 0)
-		return err;
+					2, &max_size))
+		return -EINVAL;
 
 	return 0;
 }
@@ -518,58 +336,69 @@ static void dhcp_client_event_notify(struct l_dhcp_client *client,
 
 static int dhcp_client_send_discover(struct l_dhcp_client *client)
 {
-	uint8_t *opt;
+	struct dhcp_message_builder builder;
 	size_t optlen = DHCP_MIN_OPTIONS_SIZE;
 	size_t len = sizeof(struct dhcp_message) + optlen;
 	L_AUTO_FREE_VAR(struct dhcp_message *, discover);
 	int err;
-	struct sockaddr_in si;
+
+	CLIENT_DEBUG("");
 
 	discover = (struct dhcp_message *) l_new(uint8_t, len);
 
-	err = client_message_init(client, discover,
-					DHCP_MESSAGE_TYPE_DISCOVER,
-					&opt, &optlen);
+	_dhcp_message_builder_init(&builder, discover, len,
+					DHCP_MESSAGE_TYPE_DISCOVER);
+
+	err = client_message_init(client, discover, &builder);
 	if (err < 0)
 		return err;
 
-	if (client->hostname) {
-		err = _dhcp_option_append(&opt, &optlen,
+	if (client->hostname)
+		if (!_dhcp_message_builder_append(&builder,
 						L_DHCP_OPTION_HOST_NAME,
 						strlen(client->hostname),
-						client->hostname);
-		if (err < 0)
-			return err;
-	}
+						client->hostname))
+			return -EINVAL;
 
-	err = _dhcp_option_append(&opt, &optlen, DHCP_OPTION_END, 0, NULL);
-	if (err < 0)
-		return err;
+	_dhcp_message_builder_finalize(&builder, &len);
 
-	len = dhcp_message_optimize(discover, opt);
+	return client->transport->l2_send(client->transport,
+					INADDR_ANY, DHCP_PORT_CLIENT,
+					INADDR_BROADCAST, DHCP_PORT_SERVER,
+					NULL,
+					discover, len);
+}
+
+static int dhcp_client_send_unicast(struct l_dhcp_client *client,
+					struct dhcp_message *request,
+					unsigned int len)
+{
+	struct sockaddr_in si;
 
 	memset(&si, 0, sizeof(si));
 	si.sin_family = AF_INET;
 	si.sin_port = L_CPU_TO_BE16(DHCP_PORT_SERVER);
-	si.sin_addr.s_addr = 0xffffffff;
-
-	return client->transport->send(client->transport, &si, discover, len);
+	si.sin_addr.s_addr = client->lease->server_address;
+	return client->transport->send(client->transport, &si, request, len);
 }
 
 static int dhcp_client_send_request(struct l_dhcp_client *client)
 {
-	uint8_t *opt;
+	struct dhcp_message_builder builder;
 	size_t optlen = DHCP_MIN_OPTIONS_SIZE;
 	size_t len = sizeof(struct dhcp_message) + optlen;
 	L_AUTO_FREE_VAR(struct dhcp_message *, request);
 	int err;
-	struct sockaddr_in si;
+
+	CLIENT_DEBUG("");
 
 	request = (struct dhcp_message *) l_new(uint8_t, len);
 
-	err = client_message_init(client, request,
-					DHCP_MESSAGE_TYPE_REQUEST,
-					&opt, &optlen);
+	_dhcp_message_builder_init(&builder, request, len,
+					DHCP_MESSAGE_TYPE_REQUEST);
+
+
+	err = client_message_init(client, request, &builder);
 	if (err < 0)
 		return err;
 
@@ -590,83 +419,256 @@ static int dhcp_client_send_request(struct l_dhcp_client *client)
 		 *
 		 * NOTE: 'SELECTING' is meant to be 'REQUESTING' in the RFC
 		 */
-		err = _dhcp_option_append(&opt, &optlen,
+		if (!_dhcp_message_builder_append(&builder,
 					L_DHCP_OPTION_SERVER_IDENTIFIER,
-					4, &client->lease->server_address);
-		if (err < 0)
-			return err;
+					4, &client->lease->server_address)) {
+			CLIENT_DEBUG("Failed to append server ID");
+			return -EINVAL;
+		}
 
-		err = _dhcp_option_append(&opt, &optlen,
+		if (!_dhcp_message_builder_append(&builder,
 					L_DHCP_OPTION_REQUESTED_IP_ADDRESS,
-					4, &client->lease->address);
-		if (err < 0)
-			return err;
+					4, &client->lease->address)) {
+			CLIENT_DEBUG("Failed to append requested IP");
+			return -EINVAL;
+		}
+
 		break;
+	case DHCP_STATE_RENEWING:
+	case DHCP_STATE_REBINDING:
+		request->ciaddr = client->lease->address;
+		break;
+
 	case DHCP_STATE_INIT:
 	case DHCP_STATE_SELECTING:
 	case DHCP_STATE_INIT_REBOOT:
 	case DHCP_STATE_REBOOTING:
 	case DHCP_STATE_BOUND:
-	case DHCP_STATE_RENEWING:
-	case DHCP_STATE_REBINDING:
 		return -EINVAL;
 	}
 
 	if (client->hostname) {
-		err = _dhcp_option_append(&opt, &optlen,
+		if (!_dhcp_message_builder_append(&builder,
 						L_DHCP_OPTION_HOST_NAME,
 						strlen(client->hostname),
-						client->hostname);
-		if (err < 0)
-			return err;
+						client->hostname)) {
+			CLIENT_DEBUG("Failed to append host name");
+			return -EINVAL;
+		}
 	}
 
-	err = _dhcp_option_append(&opt, &optlen, DHCP_OPTION_END, 0, NULL);
-	if (err < 0)
-		return err;
+	_dhcp_message_builder_finalize(&builder, &len);
 
-	len = dhcp_message_optimize(request, opt);
+	/*
+	 * RFC2131, Section 4.1:
+	 * "DHCP clients MUST use the IP address provided in the
+	 * 'server identifier' option for any unicast requests to the DHCP
+	 * server.
+	 */
+	if (client->state == DHCP_STATE_RENEWING)
+		return dhcp_client_send_unicast(client, request, len);
+
+	return client->transport->l2_send(client->transport,
+					INADDR_ANY, DHCP_PORT_CLIENT,
+					INADDR_BROADCAST, DHCP_PORT_SERVER,
+					NULL, request, len);
+}
+
+static void dhcp_client_send_release(struct l_dhcp_client *client)
+{
+	struct dhcp_message_builder builder;
+	size_t optlen = DHCP_MIN_OPTIONS_SIZE;
+	size_t len = sizeof(struct dhcp_message) + optlen;
+	L_AUTO_FREE_VAR(struct dhcp_message *, request);
+	int err;
+	struct sockaddr_in si;
+
+	CLIENT_DEBUG("");
 
 	memset(&si, 0, sizeof(si));
 	si.sin_family = AF_INET;
 	si.sin_port = L_CPU_TO_BE16(DHCP_PORT_SERVER);
-	si.sin_addr.s_addr = 0xffffffff;
+	si.sin_addr.s_addr = client->lease->server_address;
 
-	return client->transport->send(client->transport, &si, request, len);
+	request = (struct dhcp_message *) l_new(uint8_t, len);
+
+	_dhcp_message_builder_init(&builder, request, len,
+					DHCP_MESSAGE_TYPE_RELEASE);
+
+	err = client_message_init(client, request, &builder);
+	if (err < 0)
+		return;
+
+	request->ciaddr = client->lease->address;
+
+	if (!_dhcp_message_builder_append(&builder,
+					L_DHCP_OPTION_SERVER_IDENTIFIER,
+					4, &client->lease->server_address)) {
+		CLIENT_DEBUG("Failed to append server ID");
+		return;
+	}
+
+	_dhcp_message_builder_finalize(&builder, &len);
+
+	dhcp_client_send_unicast(client, request, len);
 }
 
 static void dhcp_client_timeout_resend(struct l_timeout *timeout,
 								void *user_data)
 {
 	struct l_dhcp_client *client = user_data;
+	unsigned int next_timeout = 0;
+	int r;
+
+	CLIENT_DEBUG("");
 
 	switch (client->state) {
-	case DHCP_STATE_INIT:
-		break;
 	case DHCP_STATE_SELECTING:
-		l_timeout_modify(timeout, 5);
-		dhcp_client_send_discover(client);
-		break;
-	case DHCP_STATE_REQUESTING:
-		if (dhcp_client_send_request(client) < 0)
+		r = dhcp_client_send_discover(client);
+		if (r < 0) {
+			CLIENT_DEBUG("Sending discover failed: %s",
+								strerror(-r));
 			goto error;
+		}
 
-		l_timeout_modify(timeout, 5);
 		break;
+	case DHCP_STATE_RENEWING:
+	case DHCP_STATE_REQUESTING:
+	case DHCP_STATE_REBINDING:
+		r = dhcp_client_send_request(client);
+		if (r < 0) {
+			CLIENT_DEBUG("Sending Request failed: %s",
+								strerror(-r));
+			goto error;
+		}
+
+		break;
+	case DHCP_STATE_INIT:
 	case DHCP_STATE_INIT_REBOOT:
 	case DHCP_STATE_REBOOTING:
 	case DHCP_STATE_BOUND:
-	case DHCP_STATE_RENEWING:
-	case DHCP_STATE_REBINDING:
 		break;
 	}
 
+	switch (client->state) {
+	case DHCP_STATE_RENEWING:
+		next_timeout = dhcp_rebind_renew_retry_time(client->start_t,
+							client->lease->t2);
+		break;
+	case DHCP_STATE_REBINDING:
+		next_timeout = dhcp_rebind_renew_retry_time(client->start_t,
+						client->lease->lifetime);
+		break;
+	case DHCP_STATE_REQUESTING:
+	case DHCP_STATE_SELECTING:
+		/*
+		 * RFC 2131 Section 4.1:
+		 * "The retransmission delay SHOULD be doubled with subsequent
+		 * retransmissions up to a maximum of 64 seconds.
+		 */
+		client->attempt += 1;
+		next_timeout = minsize(2 << client->attempt, 64);
+		break;
+	case DHCP_STATE_INIT:
+	case DHCP_STATE_INIT_REBOOT:
+	case DHCP_STATE_REBOOTING:
+	case DHCP_STATE_BOUND:
+		break;
+	}
+
+	if (next_timeout)
+		l_timeout_modify_ms(timeout, dhcp_fuzz_secs(next_timeout));
+
 	return;
 
 error:
 	l_dhcp_client_stop(client);
 }
 
+static void dhcp_client_lease_expired(struct l_timeout *timeout,
+							void *user_data)
+{
+	struct l_dhcp_client *client = user_data;
+
+	CLIENT_DEBUG("");
+
+	l_dhcp_client_stop(client);
+	dhcp_client_event_notify(client, L_DHCP_CLIENT_EVENT_LEASE_EXPIRED);
+}
+
+static void dhcp_client_t2_expired(struct l_timeout *timeout, void *user_data)
+{
+	struct l_dhcp_client *client = user_data;
+	uint32_t next_timeout = client->lease->lifetime - client->lease->t2;
+
+	CLIENT_DEBUG("");
+
+	/*
+	 * If we got here, then resend_timeout is active, with a timeout
+	 * set originally for ~60 seconds.  So we simply set the new state
+	 * and wait for the timer to fire
+	 */
+	CLIENT_ENTER_STATE(DHCP_STATE_REBINDING);
+
+	l_timeout_modify_ms(client->timeout_lease,
+				dhcp_fuzz_secs(next_timeout));
+	l_timeout_set_callback(client->timeout_lease,
+				dhcp_client_lease_expired, client, NULL);
+}
+
+static void dhcp_client_t1_expired(struct l_timeout *timeout, void *user_data)
+{
+	struct l_dhcp_client *client = user_data;
+	uint32_t next_timeout;
+	int r;
+
+	CLIENT_DEBUG("");
+
+	CLIENT_ENTER_STATE(DHCP_STATE_RENEWING);
+	client->attempt = 1;
+
+	r = dhcp_client_send_request(client);
+	if  (r < 0) {
+		CLIENT_DEBUG("Sending request failed: %s", strerror(-r));
+		goto error;
+	}
+
+	next_timeout = client->lease->t2 - client->lease->t1;
+	l_timeout_modify_ms(client->timeout_lease,
+						dhcp_fuzz_secs(next_timeout));
+	l_timeout_set_callback(client->timeout_lease, dhcp_client_t2_expired,
+				client, NULL);
+
+	next_timeout = dhcp_rebind_renew_retry_time(client->start_t,
+							client->lease->t2);
+	client->timeout_resend =
+		l_timeout_create_ms(dhcp_fuzz_secs(next_timeout),
+					dhcp_client_timeout_resend,
+					client, NULL);
+	return;
+
+error:
+	l_dhcp_client_stop(client);
+}
+
+static void dhcp_client_address_add_cb(int error, uint16_t type,
+						const void *data, uint32_t len,
+						void *user_data)
+{
+	struct l_dhcp_client *client = user_data;
+
+	client->rtnl_add_cmdid = 0;
+
+	if (error < 0 && error != -EEXIST) {
+		l_rtnl_address_free(client->rtnl_configured_address);
+		client->rtnl_configured_address = NULL;
+		CLIENT_DEBUG("Unable to set address on ifindex: %u: %d(%s)",
+				client->ifindex, error,
+				strerror(-error));
+		return;
+	}
+}
+
 static int dhcp_client_receive_ack(struct l_dhcp_client *client,
 					const struct dhcp_message *ack,
 					size_t len)
@@ -675,6 +677,8 @@ static int dhcp_client_receive_ack(struct l_dhcp_client *client,
 	struct l_dhcp_lease *lease;
 	int r;
 
+	CLIENT_DEBUG("");
+
 	if (ack->yiaddr == 0)
 		return -ENOMSG;
 
@@ -682,8 +686,11 @@ static int dhcp_client_receive_ack(struct l_dhcp_client *client,
 		return -EINVAL;
 
 	lease = _dhcp_lease_parse_options(&iter);
-	if (!lease)
+	if (!lease) {
+		CLIENT_DEBUG("Failed to parse DHCP options.");
+
 		return -ENOMSG;
+	}
 
 	lease->address = ack->yiaddr;
 
@@ -705,6 +712,38 @@ static int dhcp_client_receive_ack(struct l_dhcp_client *client,
 			client->state == DHCP_STATE_REBOOTING)
 		r = L_DHCP_CLIENT_EVENT_LEASE_OBTAINED;
 
+	if (client->rtnl) {
+		struct l_rtnl_address *a;
+		L_AUTO_FREE_VAR(char *, ip) =
+			l_dhcp_lease_get_address(client->lease);
+		uint8_t prefix_len;
+		uint32_t l = l_dhcp_lease_get_lifetime(client->lease);
+		L_AUTO_FREE_VAR(char *, netmask) =
+				l_dhcp_lease_get_netmask(client->lease);
+		L_AUTO_FREE_VAR(char *, broadcast) =
+				l_dhcp_lease_get_broadcast(client->lease);
+		struct in_addr in_addr;
+
+		if (inet_pton(AF_INET, netmask, &in_addr) > 0)
+			prefix_len = __builtin_popcountl(in_addr.s_addr);
+		else
+			prefix_len = 24;
+
+		a = l_rtnl_address_new(ip, prefix_len);
+		l_rtnl_address_set_noprefixroute(a, true);
+		l_rtnl_address_set_lifetimes(a, l, l);
+		l_rtnl_address_set_broadcast(a, broadcast);
+
+		client->rtnl_add_cmdid =
+			l_rtnl_ifaddr_add(client->rtnl, client->ifindex, a,
+						dhcp_client_address_add_cb,
+						client, NULL);
+		if (client->rtnl_add_cmdid)
+			client->rtnl_configured_address = a;
+		else
+			CLIENT_DEBUG("Configuring address via RTNL failed");
+	}
+
 	return r;
 }
 
@@ -714,6 +753,8 @@ static int dhcp_client_receive_offer(struct l_dhcp_client *client,
 {
 	struct dhcp_message_iter iter;
 
+	CLIENT_DEBUG("");
+
 	if (offer->yiaddr == 0)
 		return -ENOMSG;
 
@@ -738,8 +779,11 @@ static void dhcp_client_rx_message(const void *data, size_t len, void *userdata)
 	uint8_t t, l;
 	const void *v;
 	int r;
+	struct in_addr ia;
 
-	if (len < sizeof(struct dhcp_message) + 4)
+	CLIENT_DEBUG("");
+
+	if (len < sizeof(struct dhcp_message))
 		return;
 
 	if (message->op != DHCP_OP_CODE_BOOTREPLY)
@@ -773,14 +817,29 @@ static void dhcp_client_rx_message(const void *data, size_t len, void *userdata)
 		if (dhcp_client_receive_offer(client, message, len) < 0)
 			return;
 
-		client->state = DHCP_STATE_REQUESTING;
+		CLIENT_ENTER_STATE(DHCP_STATE_REQUESTING);
+		client->attempt = 1;
 
-		if (dhcp_client_send_request(client) < 0)
-			goto error;
+		if (dhcp_client_send_request(client) < 0) {
+			l_dhcp_client_stop(client);
 
-		l_timeout_modify(client->timeout_resend, 5);
+			return;
+		}
+
+		l_timeout_modify_ms(client->timeout_resend, dhcp_fuzz_secs(4));
 		break;
 	case DHCP_STATE_REQUESTING:
+	case DHCP_STATE_RENEWING:
+	case DHCP_STATE_REBINDING:
+		if (msg_type == DHCP_MESSAGE_TYPE_NAK) {
+			CLIENT_DEBUG("Received NAK, Stopping...");
+			l_dhcp_client_stop(client);
+
+			dhcp_client_event_notify(client,
+					L_DHCP_CLIENT_EVENT_NO_LEASE);
+			return;
+		}
+
 		if (msg_type != DHCP_MESSAGE_TYPE_ACK)
 			return;
 
@@ -788,24 +847,80 @@ static void dhcp_client_rx_message(const void *data, size_t len, void *userdata)
 		if (r < 0)
 			return;
 
-		client->state = DHCP_STATE_BOUND;
+		CLIENT_ENTER_STATE(DHCP_STATE_BOUND);
 		l_timeout_remove(client->timeout_resend);
 		client->timeout_resend = NULL;
 
+		if (client->transport->bind)
+			client->transport->bind(client->transport,
+						client->lease->address);
+
 		dhcp_client_event_notify(client, r);
+
+		/*
+		 * Start T1, once it expires we will start the T2 timer.  If
+		 * we renew the lease, we will end up back here.
+		 *
+		 * RFC2131, Section 4.4.5 states:
+		 * "Times T1 and T2 SHOULD be chosen with some random "fuzz"
+		 * around a fixed value, to avoid synchronization of client
+		 * reacquisition."
+		 */
+		l_timeout_remove(client->timeout_lease);
+		client->timeout_lease = NULL;
+
+		/* Infinite lease, no need to start t1 */
+		if (client->lease->lifetime != 0xffffffffu) {
+			uint32_t next_timeout =
+					dhcp_fuzz_secs(client->lease->t1);
+
+			CLIENT_DEBUG("T1 expiring in %u ms", next_timeout);
+			client->timeout_lease =
+				l_timeout_create_ms(next_timeout,
+							dhcp_client_t1_expired,
+							client, NULL);
+		}
+
+		/* ACD is already running, no need to re-announce */
+		if (client->acd)
+			break;
+
+		client->acd = l_acd_new(client->ifindex);
+
+		if (client->debug_handler)
+			l_acd_set_debug(client->acd, client->debug_handler,
+					client->debug_data,
+					client->debug_destroy);
+
+		/*
+		 * TODO: There is no mechanism yet to deal with IPs leased by
+		 * the DHCP server which conflict with other devices. For now
+		 * the ACD object is being initialized to defend infinitely
+		 * which is effectively no different than the non-ACD behavior
+		 * (ignore conflicts and continue using address). The only
+		 * change is that announcements will be sent if conflicts are
+		 * found.
+		 */
+		l_acd_set_defend_policy(client->acd,
+						L_ACD_DEFEND_POLICY_INFINITE);
+		l_acd_set_skip_probes(client->acd, true);
+
+		ia.s_addr = client->lease->address;
+
+		/* For unit testing we don't want this to be a fatal error */
+		if (!l_acd_start(client->acd, inet_ntoa(ia))) {
+			CLIENT_DEBUG("Failed to start ACD on %s, continuing",
+						inet_ntoa(ia));
+			l_acd_destroy(client->acd);
+			client->acd = NULL;
+		}
+
 		break;
 	case DHCP_STATE_INIT_REBOOT:
 	case DHCP_STATE_REBOOTING:
 	case DHCP_STATE_BOUND:
-	case DHCP_STATE_RENEWING:
-	case DHCP_STATE_REBINDING:
 		break;
 	}
-
-	return;
-
-error:
-	l_dhcp_client_stop(client);
 }
 
 LIB_EXPORT struct l_dhcp_client *l_dhcp_client_new(uint32_t ifindex)
@@ -833,7 +948,8 @@ LIB_EXPORT void l_dhcp_client_destroy(struct l_dhcp_client *client)
 	if (unlikely(!client))
 		return;
 
-	l_dhcp_client_stop(client);
+	if (client->state != DHCP_STATE_INIT)
+		l_dhcp_client_stop(client);
 
 	if (client->event_destroy)
 		client->event_destroy(client->event_data);
@@ -947,6 +1063,15 @@ bool _dhcp_client_set_transport(struct l_dhcp_client *client,
 	return true;
 }
 
+struct dhcp_transport *_dhcp_client_get_transport(struct l_dhcp_client *client)
+{
+	if (unlikely(!client))
+		return NULL;
+
+	return client->transport;
+}
+
+
 void _dhcp_client_override_xid(struct l_dhcp_client *client, uint32_t xid)
 {
 	client->override_xid = true;
@@ -990,33 +1115,38 @@ LIB_EXPORT bool l_dhcp_client_start(struct l_dhcp_client *client)
 
 	if (!client->transport) {
 		client->transport =
-			_dhcp_default_transport_new();
+			_dhcp_default_transport_new(client->ifindex,
+							client->ifname,
+							DHCP_PORT_CLIENT);
 
 		if (!client->transport)
 			return false;
 	}
 
+	if (!client->override_xid)
+		l_getrandom(&client->xid, sizeof(client->xid));
+
 	if (client->transport->open)
-		if (client->transport->open(client->transport, client->ifindex,
-					client->ifname, DHCP_PORT_CLIENT) < 0)
+		if (client->transport->open(client->transport,
+							client->xid) < 0)
 			return false;
 
 	_dhcp_transport_set_rx_callback(client->transport,
 						dhcp_client_rx_message,
 						client);
 
-	if (!client->override_xid)
-		l_getrandom(&client->xid, sizeof(client->xid));
-
-	client->start_t = time(NULL);
+	client->start_t = l_time_now();
 
 	err = dhcp_client_send_discover(client);
 	if (err < 0)
 		return false;
 
-	client->timeout_resend = l_timeout_create(5, dhcp_client_timeout_resend,
-							client, NULL);
-	client->state = DHCP_STATE_SELECTING;
+	client->timeout_resend = l_timeout_create_ms(dhcp_fuzz_msecs(600),
+						dhcp_client_timeout_resend,
+						client, NULL);
+	CLIENT_ENTER_STATE(DHCP_STATE_SELECTING);
+	client->attempt = 1;
+
 	return true;
 }
 
@@ -1025,15 +1155,50 @@ LIB_EXPORT bool l_dhcp_client_stop(struct l_dhcp_client *client)
 	if (unlikely(!client))
 		return false;
 
+	/*
+	 * RFC 2131 Section 4.4.6
+	 * "If the client no longer requires use of its assigned network address
+	 * (e.g., the client is gracefully shut down), the client sends a
+	 * DHCPRELEASE message to the server.""
+	 */
+	if (client->state == DHCP_STATE_BOUND ||
+			client->state == DHCP_STATE_RENEWING ||
+			client->state == DHCP_STATE_REBINDING)
+		dhcp_client_send_release(client);
+
+	if (client->rtnl_add_cmdid) {
+		l_netlink_cancel(client->rtnl, client->rtnl_add_cmdid);
+		client->rtnl_add_cmdid = 0;
+	}
+
+	if (client->rtnl_configured_address) {
+		l_rtnl_ifaddr_delete(client->rtnl, client->ifindex,
+					client->rtnl_configured_address,
+					NULL, NULL, NULL);
+		l_rtnl_address_free(client->rtnl_configured_address);
+		client->rtnl_configured_address = NULL;
+	}
+
 	l_timeout_remove(client->timeout_resend);
 	client->timeout_resend = NULL;
 
+	l_timeout_remove(client->timeout_lease);
+	client->timeout_lease = NULL;
+
 	if (client->transport && client->transport->close)
 		client->transport->close(client->transport);
 
-	client->state = DHCP_STATE_INIT;
+	client->start_t = 0;
+	CLIENT_ENTER_STATE(DHCP_STATE_INIT);
 
 	_dhcp_lease_free(client->lease);
+	client->lease = NULL;
+
+	if (client->acd) {
+		l_acd_destroy(client->acd);
+		client->acd = NULL;
+	}
+
 	return true;
 }
 
@@ -1054,3 +1219,34 @@ LIB_EXPORT bool l_dhcp_client_set_event_handler(struct l_dhcp_client *client,
 
 	return true;
 }
+
+LIB_EXPORT bool l_dhcp_client_set_debug(struct l_dhcp_client *client,
+						l_dhcp_debug_cb_t function,
+						void *user_data,
+						l_dhcp_destroy_cb_t destroy)
+{
+	if (unlikely(!client))
+		return false;
+
+	if (client->debug_destroy)
+		client->debug_destroy(client->debug_data);
+
+	client->debug_handler = function;
+	client->debug_destroy = destroy;
+	client->debug_data = user_data;
+
+	return true;
+}
+
+LIB_EXPORT bool l_dhcp_client_set_rtnl(struct l_dhcp_client *client,
+					struct l_netlink *rtnl)
+{
+	if (unlikely(!client))
+		return false;
+
+	if (unlikely(client->state != DHCP_STATE_INIT))
+		return false;
+
+	client->rtnl = rtnl;
+	return true;
+}
