@@ -24,21 +24,26 @@
 #include <config.h>
 #endif
 
+#include <arpa/inet.h>
 #include <netinet/in.h>
 #include <linux/types.h>
 #include <linux/if_arp.h>
 #include <errno.h>
 #include <time.h>
 
-#include "ell/log.h"
-#include "ell/random.h"
-#include "ell/time.h"
-#include "ell/net.h"
-#include "ell/timeout.h"
-#include "ell/uintset.h"
-#include "ell/private.h"
-#include "ell/dhcp6-private.h"
-#include "ell/dhcp6.h"
+#include "log.h"
+#include "random.h"
+#include "time.h"
+#include "time-private.h"
+#include "net.h"
+#include "timeout.h"
+#include "uintset.h"
+#include "private.h"
+#include "icmp6.h"
+#include "netlink.h"
+#include "rtnl.h"
+#include "dhcp6-private.h"
+#include "dhcp6.h"
 
 #define CLIENT_DEBUG(fmt, args...)					\
 	l_util_debug(client->debug_handler, client->debug_data,		\
@@ -58,6 +63,8 @@
 #define REQ_MAX_RT	30
 #define REN_TIMEOUT	10
 #define REN_MAX_RT	600
+#define REB_TIMEOUT	10
+#define REB_MAX_RT	600
 #define REL_TIMEOUT	1
 #define REL_MAX_RC	4
 
@@ -116,31 +123,31 @@ static const char *message_type_to_string(uint8_t t)
 static const char *option_to_string(uint16_t o)
 {
 	switch (o) {
-	case L_DHCP6_OPTION_CLIENT_ID:
+	case DHCP6_OPTION_CLIENT_ID:
 		return "CLIENTID";
-	case L_DHCP6_OPTION_SERVER_ID:
+	case DHCP6_OPTION_SERVER_ID:
 		return "SERVERID";
-	case L_DHCP6_OPTION_IA_NA:
+	case DHCP6_OPTION_IA_NA:
 		return "IA_NA";
-	case L_DHCP6_OPTION_IA_TA:
+	case DHCP6_OPTION_IA_TA:
 		return "IA_TA";
-	case L_DHCP6_OPTION_IA_PD:
+	case DHCP6_OPTION_IA_PD:
 		return "IA_PD";
-	case L_DHCP6_OPTION_REQUEST_OPTION:
+	case DHCP6_OPTION_REQUEST_OPTION:
 		return "ORO";
-	case L_DHCP6_OPTION_ELAPSED_TIME:
+	case DHCP6_OPTION_ELAPSED_TIME:
 		return "ELAPSED_TIME";
-	case L_DHCP6_OPTION_PREFERENCE:
+	case DHCP6_OPTION_PREFERENCE:
 		return "PREFERENCE";
-	case L_DHCP6_OPTION_STATUS_CODE:
+	case DHCP6_OPTION_STATUS_CODE:
 		return "STATUS_CODE";
-	case L_DHCP6_OPTION_RAPID_COMMIT:
+	case DHCP6_OPTION_RAPID_COMMIT:
 		return "RAPID_COMMIT";
-	case L_DHCP6_OPTION_USER_CLASS:
+	case DHCP6_OPTION_USER_CLASS:
 		return "USER_CLASS";
-	case L_DHCP6_OPTION_VENDOR_CLASS:
+	case DHCP6_OPTION_VENDOR_CLASS:
 		return "VENDOR_CLASS";
-	case L_DHCP6_OPTION_VENDOR_OPTS:
+	case DHCP6_OPTION_VENDOR_OPTS:
 		return "VENDOR_OPTS";
 	case L_DHCP6_OPTION_DNS_SERVERS:
 		return "DNS_SERVERS";
@@ -148,13 +155,13 @@ static const char *option_to_string(uint16_t o)
 		return "DOMAIN_LIST";
 	case L_DHCP6_OPTION_SNTP_SERVERS:
 		return "SNTP_SERVERS";
-	case L_DHCP6_OPTION_INF_RT:
+	case DHCP6_OPTION_INF_RT:
 		return "INF_RT";
 	case L_DHCP6_OPTION_NTP_SERVER:
 		return "NTP_SERVER";
-	case L_DHCP6_OPTION_SOL_MAX_RT:
+	case DHCP6_OPTION_SOL_MAX_RT:
 		return "SOL_MAX_RT";
-	case L_DHCP6_OPTION_INF_MAX_RT:
+	case DHCP6_OPTION_INF_MAX_RT:
 		return "INF_MAX_RT";
 	default:
 		break;
@@ -224,8 +231,7 @@ static uint8_t *option_reserve(struct dhcp6_message_builder *builder,
 	return p;
 }
 
-static void option_start(struct dhcp6_message_builder *builder,
-						enum l_dhcp6_option type)
+static void option_start(struct dhcp6_message_builder *builder, uint16_t type)
 {
 	builder->option_start = builder->options_pos;
 	l_put_be16(type, option_reserve(builder, OPTION_HEADER_LEN));
@@ -242,7 +248,7 @@ static void option_finalize(struct dhcp6_message_builder *builder)
 }
 
 static void option_append_uint16(struct dhcp6_message_builder *builder,
-					enum l_dhcp6_option type, uint16_t v)
+					uint16_t type, uint16_t v)
 {
 	uint8_t *p = option_reserve(builder, OPTION_HEADER_LEN + sizeof(v));
 
@@ -252,8 +258,7 @@ static void option_append_uint16(struct dhcp6_message_builder *builder,
 }
 
 static void option_append_bytes(struct dhcp6_message_builder *builder,
-				enum l_dhcp6_option type,
-				const void *bytes, uint16_t len)
+				uint16_t type, const void *bytes, uint16_t len)
 {
 	uint8_t *p = option_reserve(builder, OPTION_HEADER_LEN + len);
 
@@ -328,7 +333,7 @@ static void option_append_elapsed_time(struct dhcp6_message_builder *builder,
 		elapsed_time = UINT16_MAX;
 
 done:
-	option_append_uint16(builder, L_DHCP6_OPTION_ELAPSED_TIME,
+	option_append_uint16(builder, DHCP6_OPTION_ELAPSED_TIME,
 							elapsed_time);
 }
 
@@ -339,6 +344,7 @@ enum dhcp6_state {
 	DHCP6_STATE_REQUESTING,
 	DHCP6_STATE_BOUND,
 	DHCP6_STATE_RENEWING,
+	DHCP6_STATE_REBINDING,
 	DHCP6_STATE_RELEASING,
 };
 
@@ -357,6 +363,8 @@ static const char *dhcp6_state_to_str(enum dhcp6_state s)
 		return "Bound";
 	case DHCP6_STATE_RENEWING:
 		return "Renewing";
+	case DHCP6_STATE_REBINDING:
+		return "Rebinding";
 	case DHCP6_STATE_RELEASING:
 		return "Releasing";
 	};
@@ -384,6 +392,14 @@ struct l_dhcp6_client {
 
 	struct l_timeout *timeout_send;
 	struct l_dhcp6_lease *lease;
+	struct l_timeout *timeout_lease;
+	uint64_t lease_start_t;
+
+	struct l_icmp6_client *icmp6;
+
+	struct l_netlink *rtnl;
+	uint32_t rtnl_add_cmdid;
+	struct l_rtnl_address *rtnl_configured_address;
 
 	l_dhcp6_client_event_cb_t event_handler;
 	void *event_data;
@@ -393,14 +409,19 @@ struct l_dhcp6_client {
 	uint8_t addr_len;
 	uint8_t addr_type;
 
+	struct in6_addr ll_address;
+
 	l_dhcp6_debug_cb_t debug_handler;
 	l_dhcp6_destroy_cb_t debug_destroy;
 	void *debug_data;
 
 	bool stateless : 1;
 	bool nodelay : 1;
+	bool nora : 1;
 	bool request_pd : 1;
 	bool request_na : 1;
+	bool no_rapid_commit : 1;
+	bool lla_randomized : 1;
 };
 
 static const struct in6_addr all_nodes = DHCP6_ADDR_LINKLOCAL_ALL_NODES;
@@ -418,22 +439,23 @@ static void option_append_option_request(struct dhcp6_message_builder *builder,
 {
 	struct l_uintset *clone = NULL;
 
-	option_start(builder, L_DHCP6_OPTION_REQUEST_OPTION);
+	option_start(builder, DHCP6_OPTION_REQUEST_OPTION);
 
 	switch (state) {
 	case DHCP6_STATE_SOLICITING:
 	case DHCP6_STATE_REQUESTING:
 		clone = l_uintset_clone(request_options);
-		l_uintset_put(clone, L_DHCP6_OPTION_SOL_MAX_RT);
+		l_uintset_put(clone, DHCP6_OPTION_SOL_MAX_RT);
 		break;
 	case DHCP6_STATE_REQUESTING_INFORMATION:
 		clone = l_uintset_clone(request_options);
-		l_uintset_put(clone, L_DHCP6_OPTION_INF_RT);
-		l_uintset_put(clone, L_DHCP6_OPTION_INF_MAX_RT);
+		l_uintset_put(clone, DHCP6_OPTION_INF_RT);
+		l_uintset_put(clone, DHCP6_OPTION_INF_MAX_RT);
 		break;
 	case DHCP6_STATE_INIT:
 	case DHCP6_STATE_BOUND:
 	case DHCP6_STATE_RENEWING:
+	case DHCP6_STATE_REBINDING:
 	case DHCP6_STATE_RELEASING:
 		break;
 	}
@@ -445,16 +467,9 @@ static void option_append_option_request(struct dhcp6_message_builder *builder,
 	l_uintset_free(clone);
 }
 
-static void option_append_reconfigure_accept(
-					struct dhcp6_message_builder *builder)
-{
-	option_start(builder, L_DHCP6_OPTION_RECONF_ACCEPT);
-	option_finalize(builder);
-}
-
 static int option_append_ia_common(struct l_dhcp6_client *client,
 				struct dhcp6_message_builder *builder,
-				enum l_dhcp6_option option)
+				uint16_t option)
 {
 	option_start(builder, option);
 
@@ -488,7 +503,7 @@ static void option_append_ia_na(struct l_dhcp6_client *client,
 {
 	void *ia_addr;
 
-	if (option_append_ia_common(client, builder, L_DHCP6_OPTION_IA_NA) < 0)
+	if (option_append_ia_common(client, builder, DHCP6_OPTION_IA_NA) < 0)
 		return;
 
 	l_put_be32(0, option_reserve(builder, 4));
@@ -498,7 +513,7 @@ static void option_append_ia_na(struct l_dhcp6_client *client,
 		goto done;
 
 	ia_addr = option_reserve(builder, 28);
-	l_put_be16(L_DHCP6_OPTION_IA_ADDR, ia_addr);
+	l_put_be16(DHCP6_OPTION_IA_ADDR, ia_addr);
 	l_put_be16(24, ia_addr + 2);
 	memcpy(ia_addr + 4, client->lease->ia_na.info.addr, 16);
 	l_put_be32(client->lease->ia_na.info.preferred_lifetime, ia_addr + 20);
@@ -511,7 +526,7 @@ done:
 static void option_append_ia_pd(struct l_dhcp6_client *client,
 				struct dhcp6_message_builder *builder)
 {
-	if (option_append_ia_common(client, builder, L_DHCP6_OPTION_IA_PD) < 0)
+	if (option_append_ia_common(client, builder, DHCP6_OPTION_IA_PD) < 0)
 		return;
 
 	l_put_be32(0, option_reserve(builder, 4));
@@ -523,40 +538,24 @@ static void option_append_ia_pd(struct l_dhcp6_client *client,
 static void client_enable_option(struct l_dhcp6_client *client,
 						enum l_dhcp6_option option)
 {
-	size_t i;
+	const char *s;
 
-	static const struct {
-		enum l_dhcp6_option option;
-	} options_to_ignore[] = {
-		{ L_DHCP6_OPTION_CLIENT_ID },
-		{ L_DHCP6_OPTION_SERVER_ID },
-		{ L_DHCP6_OPTION_IA_NA },
-		{ L_DHCP6_OPTION_IA_TA },
-		{ L_DHCP6_OPTION_IA_PD },
-		{ L_DHCP6_OPTION_IA_ADDR },
-		{ L_DHCP6_OPTION_IA_PREFIX },
-		{ L_DHCP6_OPTION_REQUEST_OPTION },
-		{ L_DHCP6_OPTION_ELAPSED_TIME },
-		{ L_DHCP6_OPTION_PREFERENCE },
-		{ L_DHCP6_OPTION_RELAY_MSG },
-		{ L_DHCP6_OPTION_AUTH },
-		{ L_DHCP6_OPTION_UNICAST },
-		{ L_DHCP6_OPTION_STATUS_CODE },
-		{ L_DHCP6_OPTION_RAPID_COMMIT },
-		{ L_DHCP6_OPTION_USER_CLASS },
-		{ L_DHCP6_OPTION_VENDOR_CLASS },
-		{ L_DHCP6_OPTION_INTERFACE_ID },
-		{ L_DHCP6_OPTION_RECONF_MSG },
-		{ L_DHCP6_OPTION_RECONF_ACCEPT },
-		{ L_DHCP6_OPTION_INF_RT },
-		{ L_DHCP6_OPTION_SOL_MAX_RT },
-		{ L_DHCP6_OPTION_INF_MAX_RT },
-		{ }
-	};
+	switch (option) {
+	case L_DHCP6_OPTION_DNS_SERVERS:
+	case L_DHCP6_OPTION_DOMAIN_LIST:
+	case L_DHCP6_OPTION_SNTP_SERVERS:
+	case L_DHCP6_OPTION_NTP_SERVER:
+		break;
+	default:
+		s = option_to_string(option);
 
-	for (i = 0; options_to_ignore[i].option; i++)
-		if (options_to_ignore[i].option == option)
-			return;
+		if (s)
+			CLIENT_DEBUG("Ignore request option: %s", s);
+		else
+			CLIENT_DEBUG("Ignore request option: %u", option);
+
+		return;
+	}
 
 	l_uintset_put(client->request_options, option);
 }
@@ -588,16 +587,23 @@ static void client_duid_generate_addr_plus_time(struct l_dhcp6_client *client)
 							client->addr_len);
 }
 
-static uint64_t fuzz_msecs(uint64_t ms)
+static void client_duid_generate_addr(struct l_dhcp6_client *client)
 {
+	if (client->duid)
+		return;
+
 	/*
-	 * Compute ms + RAND*ms where RAND is in range -0.1 .. 0.1
-	 *
-	 * We do this by subtracting 0.1ms and adding 0.1ms * rand[0 .. 2]
+	 * RFC 7844, Section 4.3 Client Identifier DHCPv6 Option:
+	 * "When using the anonymity profile in conjunction with randomized
+	 * link-layer addresses, DHCPv6 clients MUST use DUID format
+	 * number 3 -- link-layer address.  The value of the link-layer
+	 * address should be the value currently assigned to the interface.
 	 */
-        return ms - ms / 10 +
-			(l_getrandom_uint32() % (2 * L_MSEC_PER_SEC)) *
-						ms / 10 / L_MSEC_PER_SEC;
+	client->duid_len = 4 + client->addr_len;
+	client->duid = l_malloc(client->duid_len);
+	client->duid->type = L_CPU_TO_BE16(DUID_TYPE_LINK_LAYER_ADDR);
+	l_put_be16(client->addr_type, client->duid->identifier);
+	memcpy(client->duid->identifier + 2, client->addr, client->addr_len);
 }
 
 static void set_retransmission_delay(struct l_dhcp6_client *client,
@@ -630,7 +636,7 @@ static void set_retransmission_delay(struct l_dhcp6_client *client,
 	 * 		RT = MRT + RAND*MRT
 	 */
 	if (!client->attempt_delay) {
-		client->attempt_delay = fuzz_msecs(irt_ms);
+		client->attempt_delay = _time_fuzz_msecs(irt_ms);
 
 		/*
 		 * RFC 8415, Section 18.2.1:
@@ -642,10 +648,10 @@ static void set_retransmission_delay(struct l_dhcp6_client *client,
 			client->attempt_delay += irt_ms / 10;
 	} else {
 		if (mrt_ms && client->attempt_delay > mrt_ms)
-			client->attempt_delay = fuzz_msecs(mrt_ms);
+			client->attempt_delay = _time_fuzz_msecs(mrt_ms);
 		else
 			client->attempt_delay +=
-					fuzz_msecs(client->attempt_delay);
+					_time_fuzz_msecs(client->attempt_delay);
 	}
 
 	l_timeout_modify_ms(client->timeout_send, client->attempt_delay);
@@ -660,11 +666,12 @@ static struct dhcp6_message *dhcp6_client_build_message(
 
 	builder = dhcp6_message_builder_new(type, client->transaction_id, 128);
 
-	option_append_bytes(builder, L_DHCP6_OPTION_CLIENT_ID,
+	option_append_bytes(builder, DHCP6_OPTION_CLIENT_ID,
 					client->duid, client->duid_len);
 
-	if (type == DHCP6_MESSAGE_TYPE_REQUEST)
-		option_append_bytes(builder, L_DHCP6_OPTION_SERVER_ID,
+	if (type == DHCP6_MESSAGE_TYPE_REQUEST ||
+			type == DHCP6_MESSAGE_TYPE_RENEW)
+		option_append_bytes(builder, DHCP6_OPTION_SERVER_ID,
 					client->lease->server_id,
 					client->lease->server_id_len);
 
@@ -676,6 +683,10 @@ static struct dhcp6_message *dhcp6_client_build_message(
 
 	option_append_elapsed_time(builder, client->transaction_start_t);
 
+	if (type == DHCP6_MESSAGE_TYPE_SOLICIT && !client->no_rapid_commit)
+		option_append_bytes(builder, DHCP6_OPTION_RAPID_COMMIT,
+						NULL, 0);
+
 	if (client->request_pd)
 		option_append_ia_pd(client, builder);
 
@@ -727,7 +738,6 @@ static int dhcp6_client_send_information_request(struct l_dhcp6_client *client)
 	option_append_elapsed_time(builder, client->transaction_start_t);
 	option_append_option_request(builder, client->request_options,
 					DHCP6_STATE_REQUESTING_INFORMATION);
-	option_append_reconfigure_accept(builder);
 
 	information_request = dhcp6_message_builder_free(builder, false,
 						&information_request_len);
@@ -740,7 +750,30 @@ static int dhcp6_client_send_information_request(struct l_dhcp6_client *client)
 
 static int dhcp6_client_send_renew(struct l_dhcp6_client *client)
 {
-	return 0;
+	L_AUTO_FREE_VAR(struct dhcp6_message *, renew);
+	size_t renew_len;
+
+	CLIENT_DEBUG("");
+
+	renew = dhcp6_client_build_message(client,
+						DHCP6_MESSAGE_TYPE_RENEW,
+						&renew_len);
+	return client->transport->send(client->transport, &all_nodes,
+						renew, renew_len);
+}
+
+static int dhcp6_client_send_rebind(struct l_dhcp6_client *client)
+{
+	L_AUTO_FREE_VAR(struct dhcp6_message *, rebind);
+	size_t rebind_len;
+
+	CLIENT_DEBUG("");
+
+	rebind = dhcp6_client_build_message(client,
+						DHCP6_MESSAGE_TYPE_REBIND,
+						&rebind_len);
+	return client->transport->send(client->transport, &all_nodes,
+						rebind, rebind_len);
 }
 
 static int dhcp6_client_send_release(struct l_dhcp6_client *client)
@@ -782,6 +815,13 @@ static int dhcp6_client_send_next(struct l_dhcp6_client *client)
 
 		set_retransmission_delay(client, REN_TIMEOUT, REN_MAX_RT, 0);
 		break;
+	case DHCP6_STATE_REBINDING:
+		r = dhcp6_client_send_rebind(client);
+		if (r < 0)
+			return r;
+
+		set_retransmission_delay(client, REB_TIMEOUT, REB_MAX_RT, 0);
+		break;
 	case DHCP6_STATE_RELEASING:
 		r = dhcp6_client_send_release(client);
 		if (r < 0)
@@ -809,15 +849,11 @@ static inline void dhcp6_client_event_notify(struct l_dhcp6_client *client,
 		client->event_handler(client, event, client->event_data);
 }
 
-static void dhcp6_client_setup_lease(struct l_dhcp6_client *client)
-{
-	dhcp6_client_event_notify(client, L_DHCP6_CLIENT_EVENT_LEASE_OBTAINED);
-}
-
 static inline void dhcp6_client_new_transaction(struct l_dhcp6_client *client,
 						enum dhcp6_state new_state)
 {
 	client->attempt = 0;
+	client->attempt_delay = 0;
 	client->transaction_id = l_getrandom_uint32() & 0x00FFFFFFU;
 	client->transaction_start_t = 0;
 
@@ -826,6 +862,151 @@ static inline void dhcp6_client_new_transaction(struct l_dhcp6_client *client,
 			"Entering state: %s", dhcp6_state_to_str(new_state));
 }
 
+static void dhcp6_client_timeout_send(struct l_timeout *timeout,
+								void *user_data)
+{
+	struct l_dhcp6_client *client = user_data;
+
+	CLIENT_DEBUG("");
+
+	if (client->state == DHCP6_STATE_SOLICITING
+			&& client->attempt && client->lease) {
+		CLIENT_DEBUG("Received a lease during initial request time");
+		dhcp6_client_new_transaction(client, DHCP6_STATE_REQUESTING);
+	}
+
+	if (dhcp6_client_send_next(client) < 0)
+		l_dhcp6_client_stop(client);
+}
+
+static void dhcp6_client_lease_expired(struct l_timeout *timeout,
+							void *user_data)
+{
+	struct l_dhcp6_client *client = user_data;
+
+	CLIENT_DEBUG("");
+
+	l_dhcp6_client_stop(client);
+	dhcp6_client_event_notify(client, L_DHCP6_CLIENT_EVENT_LEASE_EXPIRED);
+}
+
+static void dhcp6_client_t2_expired(struct l_timeout *timeout, void *user_data)
+{
+	struct l_dhcp6_client *client = user_data;
+	uint32_t elapsed =
+		l_time_to_msecs(l_time_now() - client->lease_start_t);
+	uint32_t valid = _dhcp6_lease_get_valid_lifetime(client->lease) *
+				L_MSEC_PER_SEC;
+
+	CLIENT_DEBUG("");
+
+	dhcp6_client_new_transaction(client, DHCP6_STATE_REBINDING);
+	l_timeout_set_callback(client->timeout_lease,
+				dhcp6_client_lease_expired, client, NULL);
+
+	if (valid <= elapsed) {
+		dhcp6_client_lease_expired(client->timeout_lease, client);
+		return;
+	}
+
+	l_timeout_modify_ms(client->timeout_lease, valid - elapsed);
+}
+
+static void dhcp6_client_t1_expired(struct l_timeout *timeout, void *user_data)
+{
+	struct l_dhcp6_client *client = user_data;
+	uint32_t next_timeout;
+
+	CLIENT_DEBUG("");
+
+	/* timeout_send was destroyed when entering BOUND, re-create */
+	client->timeout_send = l_timeout_create_ms(0,
+						dhcp6_client_timeout_send,
+						client, NULL);
+	dhcp6_client_new_transaction(client, DHCP6_STATE_RENEWING);
+
+	if (dhcp6_client_send_next(client) < 0) {
+		l_dhcp6_client_stop(client);
+		return;
+	}
+
+	next_timeout = _dhcp6_lease_get_t2(client->lease) -
+					_dhcp6_lease_get_t1(client->lease);
+	l_timeout_modify(client->timeout_lease, next_timeout);
+	l_timeout_set_callback(client->timeout_lease, dhcp6_client_t2_expired,
+					client, NULL);
+}
+
+static void dhcp6_client_address_add_cb(int error, uint16_t type,
+						const void *data, uint32_t len,
+						void *user_data)
+{
+	struct l_dhcp6_client *client = user_data;
+
+	client->rtnl_add_cmdid = 0;
+
+	if (error < 0 && error != -EEXIST) {
+		l_rtnl_address_free(client->rtnl_configured_address);
+		client->rtnl_configured_address = NULL;
+		CLIENT_DEBUG("Unable to set address on ifindex: %u: %d(%s)",
+				client->ifindex, error,
+				strerror(-error));
+		return;
+	}
+}
+
+static void dhcp6_client_setup_lease(struct l_dhcp6_client *client)
+{
+	uint32_t t1 = _dhcp6_lease_get_t1(client->lease);
+	uint32_t t2 = _dhcp6_lease_get_t2(client->lease);
+
+	client->lease_start_t = l_time_now();
+
+	/* TODO: Emit IP_CHANGED if any addresses were removed / added */
+	if (client->state == DHCP6_STATE_REQUESTING ||
+			client->state == DHCP6_STATE_SOLICITING)
+		dhcp6_client_event_notify(client,
+					L_DHCP6_CLIENT_EVENT_LEASE_OBTAINED);
+	else
+		dhcp6_client_event_notify(client,
+					L_DHCP6_CLIENT_EVENT_LEASE_RENEWED);
+
+	l_timeout_remove(client->timeout_lease);
+	client->timeout_lease = NULL;
+
+	if (t1 == 0xffffffff || t2 == 0xffffffff) {
+		CLIENT_DEBUG("T1 (%u) or T2 (%u) was infinite", t1, t2);
+		return;
+	}
+
+	client->timeout_lease = l_timeout_create_ms(_time_fuzz_secs(t1, 1),
+							dhcp6_client_t1_expired,
+							client, NULL);
+
+	if (client->rtnl) {
+		struct l_rtnl_address *a;
+		L_AUTO_FREE_VAR(char *, ip) =
+			l_dhcp6_lease_get_address(client->lease);
+		uint8_t prefix_len =
+			l_dhcp6_lease_get_prefix_length(client->lease);
+		uint32_t p = _dhcp6_lease_get_preferred_lifetime(client->lease);
+		uint32_t v = _dhcp6_lease_get_valid_lifetime(client->lease);
+
+		a = l_rtnl_address_new(ip, prefix_len);
+		l_rtnl_address_set_noprefixroute(a, true);
+		l_rtnl_address_set_lifetimes(a, p, v);
+
+		client->rtnl_add_cmdid =
+			l_rtnl_ifaddr_add(client->rtnl, client->ifindex, a,
+						dhcp6_client_address_add_cb,
+						client, NULL);
+		if (client->rtnl_add_cmdid)
+			client->rtnl_configured_address = a;
+		else
+			CLIENT_DEBUG("Configuring address via RTNL failed");
+	}
+}
+
 void __dhcp6_option_iter_init(struct dhcp6_option_iter *iter,
 				const void *options, size_t len)
 {
@@ -928,27 +1109,27 @@ static int dhcp6_client_validate_message(struct l_dhcp6_client *client,
 		 * singletons
 		 */
 		switch (t) {
-		case L_DHCP6_OPTION_CLIENT_ID:
-		case L_DHCP6_OPTION_SERVER_ID:
-		case L_DHCP6_OPTION_IA_NA:
-		case L_DHCP6_OPTION_IA_TA:
-		case L_DHCP6_OPTION_IA_PD:
-		case L_DHCP6_OPTION_REQUEST_OPTION:
-		case L_DHCP6_OPTION_ELAPSED_TIME:
-		case L_DHCP6_OPTION_PREFERENCE:
-		case L_DHCP6_OPTION_RAPID_COMMIT:
-		case L_DHCP6_OPTION_USER_CLASS:
+		case DHCP6_OPTION_CLIENT_ID:
+		case DHCP6_OPTION_SERVER_ID:
+		case DHCP6_OPTION_IA_NA:
+		case DHCP6_OPTION_IA_TA:
+		case DHCP6_OPTION_IA_PD:
+		case DHCP6_OPTION_REQUEST_OPTION:
+		case DHCP6_OPTION_ELAPSED_TIME:
+		case DHCP6_OPTION_PREFERENCE:
+		case DHCP6_OPTION_RAPID_COMMIT:
+		case DHCP6_OPTION_USER_CLASS:
 		case L_DHCP6_OPTION_DNS_SERVERS:
 		case L_DHCP6_OPTION_DOMAIN_LIST:
 		case L_DHCP6_OPTION_SNTP_SERVERS:
-		case L_DHCP6_OPTION_INF_RT:
+		case DHCP6_OPTION_INF_RT:
 		case L_DHCP6_OPTION_NTP_SERVER:
-		case L_DHCP6_OPTION_SOL_MAX_RT:
-		case L_DHCP6_OPTION_INF_MAX_RT:
+		case DHCP6_OPTION_SOL_MAX_RT:
+		case DHCP6_OPTION_INF_MAX_RT:
 			break;
-		case L_DHCP6_OPTION_STATUS_CODE:
-		case L_DHCP6_OPTION_VENDOR_CLASS:
-		case L_DHCP6_OPTION_VENDOR_OPTS:
+		case DHCP6_OPTION_STATUS_CODE:
+		case DHCP6_OPTION_VENDOR_CLASS:
+		case DHCP6_OPTION_VENDOR_OPTS:
 		default:
 			/* Multiples allowed */
 			continue;
@@ -978,7 +1159,7 @@ static int dhcp6_client_validate_message(struct l_dhcp6_client *client,
 
 	while (_dhcp6_option_iter_next(&iter, &t, &l, &v)) {
 		switch (t) {
-		case L_DHCP6_OPTION_CLIENT_ID:
+		case DHCP6_OPTION_CLIENT_ID:
 			if (client->duid_len != l ||
 					memcmp(client->duid, v, l)) {
 				CLIENT_DEBUG("Message %s - invalid option: %s",
@@ -988,7 +1169,7 @@ static int dhcp6_client_validate_message(struct l_dhcp6_client *client,
 
 			duid_verified = true;
 			break;
-		case L_DHCP6_OPTION_STATUS_CODE:
+		case DHCP6_OPTION_STATUS_CODE:
 			if (l < 2)
 				return -EBADMSG;
 
@@ -998,7 +1179,7 @@ static int dhcp6_client_validate_message(struct l_dhcp6_client *client,
 				return -EINVAL;
 			}
 			break;
-		case L_DHCP6_OPTION_SERVER_ID:
+		case DHCP6_OPTION_SERVER_ID:
 			server_id_present = true;
 			break;
 		default:
@@ -1027,6 +1208,9 @@ static int dhcp6_client_receive_advertise(struct l_dhcp6_client *client,
 	struct l_dhcp6_lease *lease;
 	struct dhcp6_option_iter iter;
 
+	if (advertise->msg_type != DHCP6_MESSAGE_TYPE_ADVERTISE)
+		return -EINVAL;
+
 	r = dhcp6_client_validate_message(client, advertise, len);
 	if (r < 0)
 		return r;
@@ -1110,6 +1294,9 @@ static int dhcp6_client_receive_reply(struct l_dhcp6_client *client,
 	struct dhcp6_option_iter iter;
 	int r;
 
+	if (reply->msg_type != DHCP6_MESSAGE_TYPE_REPLY)
+		return -EINVAL;
+
 	r = dhcp6_client_validate_message(client, reply, len);
 	if (r < 0)
 		return r;
@@ -1131,6 +1318,13 @@ static int dhcp6_client_receive_reply(struct l_dhcp6_client *client,
 		goto bad_lease;
 	}
 
+	if (client->state == DHCP6_STATE_SOLICITING && !lease->rapid_commit) {
+		CLIENT_DEBUG("Did not request Rapid Commit, but received one"
+				" anyway.  Ignoring");
+		_dhcp6_lease_free(client->lease);
+		return -EBADMSG;
+	}
+
 	_dhcp6_lease_free(client->lease);
 	client->lease = lease;
 
@@ -1173,71 +1367,133 @@ static void dhcp6_client_rx_message(const void *data, size_t len,
 	case DHCP6_STATE_INIT:
 	case DHCP6_STATE_BOUND:
 		return;
-	case DHCP6_STATE_SOLICITING:
-		if (message->msg_type != DHCP6_MESSAGE_TYPE_ADVERTISE)
-			return;
-	
-		r = dhcp6_client_receive_advertise(client, message, len);
-		if (r < 0)
-			return;
-
-		break;
 	case DHCP6_STATE_REQUESTING_INFORMATION:
-		if (message->msg_type != DHCP6_MESSAGE_TYPE_REPLY)
-			return;
-
 		if (dhcp6_client_receive_reply(client, message, len) < 0)
 			return;
 
 		break;
-	case DHCP6_STATE_REQUESTING:
-		if (message->msg_type != DHCP6_MESSAGE_TYPE_REPLY)
+	case DHCP6_STATE_SOLICITING:
+		r = dhcp6_client_receive_advertise(client, message, len);
+		if (r >= 0)
+			break;
+
+		if (client->no_rapid_commit)
 			return;
 
+		/* Fall through */
+	case DHCP6_STATE_REQUESTING:
+	case DHCP6_STATE_RENEWING:
+	case DHCP6_STATE_REBINDING:
 		r = dhcp6_client_receive_reply(client, message, len);
 		if (r < 0)
 			return;
 		break;
-	case DHCP6_STATE_RENEWING:
-		if (message->msg_type != DHCP6_MESSAGE_TYPE_REPLY)
-			return;
-
-		break;
 	case DHCP6_STATE_RELEASING:
-		if (message->msg_type != DHCP6_MESSAGE_TYPE_REPLY)
-			return;
 		break;
 	}
 
 	if (r == (int) client->state)
 		return;
 
-	dhcp6_client_new_transaction(client, r);
-
-	if (client->state == DHCP6_STATE_BOUND) {
+	if (r == DHCP6_STATE_BOUND) {
+		l_timeout_remove(client->timeout_send);
+		client->timeout_send = NULL;
 		dhcp6_client_setup_lease(client);
+		dhcp6_client_new_transaction(client, r);
 		return;
 	}
 
+	dhcp6_client_new_transaction(client, r);
+
 	if (dhcp6_client_send_next(client) < 0)
 		l_dhcp6_client_stop(client);
 }
 
-static void dhcp6_client_timeout_send(struct l_timeout *timeout,
-								void *user_data)
+static void dhcp6_client_send_initial(struct l_dhcp6_client *client)
+{
+	uint32_t delay;
+
+	if (client->stateless) {
+		dhcp6_client_new_transaction(client,
+					DHCP6_STATE_REQUESTING_INFORMATION);
+		delay = _time_pick_interval_secs(0, INF_MAX_DELAY);
+	} else {
+		dhcp6_client_new_transaction(client, DHCP6_STATE_SOLICITING);
+		delay = _time_pick_interval_secs(0, SOL_MAX_DELAY);
+	}
+
+	/*
+	 * RFC 8415, Section 18.2.1:
+	 * "The first Solicit message from the client on the interface SHOULD
+	 * be delayed by a random amount of time between 0 and SOL_MAX_DELAY.
+	 * This random delay helps desynchronize clients that start a DHCP
+	 * session at the same time, such as after recovery from a power
+	 * failure or after a router outage after seeing that DHCP is
+	 * available in Router Advertisement messages..."
+	 *
+	 * Similar verbiage in Section 18.2.6, except it uses 'MUST':
+	 * "The first Information-request message from the client on the
+	 * interface MUST be delayed by a random amount of time between 0 and
+	 * INF_MAX_DELAY."
+	 *
+	 * In certain situations, like when a WiFi network was just joined,
+	 * there's no point in waiting a random time, so we forgo the wait
+	 * if no delay is requested
+	 */
+	if (client->nodelay)
+		delay = 0;
+
+	client->timeout_send = l_timeout_create_ms(delay,
+						dhcp6_client_timeout_send,
+						client, NULL);
+
+	if (client->nodelay)
+		dhcp6_client_timeout_send(NULL, client);
+}
+
+static void dhcp6_client_icmp6_event(struct l_icmp6_client *icmp6,
+					enum l_icmp6_client_event event,
+					void *user_data)
 {
 	struct l_dhcp6_client *client = user_data;
 
-	CLIENT_DEBUG("");
+	l_timeout_remove(client->timeout_send);
+	client->timeout_send = NULL;
 
-	if (client->state == DHCP6_STATE_SOLICITING
-			&& client->attempt && client->lease) {
-		CLIENT_DEBUG("Received a lease during initial request time");
-		dhcp6_client_new_transaction(client, DHCP6_STATE_REQUESTING);
+	switch (event) {
+	case L_ICMP6_CLIENT_EVENT_ROUTER_FOUND:
+	{
+		const struct l_icmp6_router *r =
+			l_icmp6_client_get_router(icmp6);
+		bool managed = l_icmp6_router_get_managed(r);
+		bool other = l_icmp6_router_get_other(r);
+
+		CLIENT_DEBUG("Received RA, managed: %s, other: %s",
+				managed ? "yes" : "no",
+				other ? "yes" : "no");
+
+		if (!managed && !other) {
+			l_dhcp6_client_stop(client);
+			dhcp6_client_event_notify(client,
+						L_DHCP6_CLIENT_EVENT_NO_LEASE);
+			return;
+		}
+
+		if (l_icmp6_router_get_managed(r))
+			dhcp6_client_send_initial(client);
+
+		break;
 	}
+	}
+}
 
-	if (dhcp6_client_send_next(client) < 0)
-		l_dhcp6_client_stop(client);
+static void dhcp6_client_no_ra(struct l_timeout *timeout, void *user_data)
+{
+	struct l_dhcp6_client *client = user_data;
+
+	CLIENT_DEBUG("No Router Advertisements received, assume no DHCPv6");
+	l_dhcp6_client_stop(client);
+	dhcp6_client_event_notify(client, L_DHCP6_CLIENT_EVENT_NO_LEASE);
 }
 
 LIB_EXPORT struct l_dhcp6_client *l_dhcp6_client_new(uint32_t ifindex)
@@ -1250,6 +1506,11 @@ LIB_EXPORT struct l_dhcp6_client *l_dhcp6_client_new(uint32_t ifindex)
 	client->ifindex = ifindex;
 	client->request_na = true;
 
+	client->icmp6 = l_icmp6_client_new(ifindex);
+	l_icmp6_client_set_event_handler(client->icmp6,
+						dhcp6_client_icmp6_event,
+						client, NULL);
+
 	client->request_options = l_uintset_new(256);
 	client_enable_option(client, L_DHCP6_OPTION_DOMAIN_LIST);
 	client_enable_option(client, L_DHCP6_OPTION_DNS_SERVERS);
@@ -1262,7 +1523,10 @@ LIB_EXPORT void l_dhcp6_client_destroy(struct l_dhcp6_client *client)
 	if (unlikely(!client))
 		return;
 
-	l_dhcp6_client_stop(client);
+	if (client->state != DHCP6_STATE_INIT)
+		l_dhcp6_client_stop(client);
+
+	l_icmp6_client_free(client->icmp6);
 
 	if (client->event_destroy)
 		client->event_destroy(client->event_data);
@@ -1270,7 +1534,6 @@ LIB_EXPORT void l_dhcp6_client_destroy(struct l_dhcp6_client *client)
 	_dhcp6_transport_free(client->transport);
 	l_uintset_free(client->request_options);
 	l_free(client->duid);
-	l_timeout_remove(client->timeout_send);
 
 	l_free(client);
 }
@@ -1289,6 +1552,7 @@ LIB_EXPORT const struct l_dhcp6_lease *l_dhcp6_client_get_lease(
 		return NULL;
 	case DHCP6_STATE_BOUND:
 	case DHCP6_STATE_RENEWING:
+	case DHCP6_STATE_REBINDING:
 	case DHCP6_STATE_REQUESTING_INFORMATION:
 		break;
 	}
@@ -1320,6 +1584,24 @@ LIB_EXPORT bool l_dhcp6_client_set_address(struct l_dhcp6_client *client,
 	return true;
 }
 
+/*
+ * Set the link local address to use instead of binding to the in6addr_any
+ * address by default.  This allows multiple clients to coexist simulatenously
+ * on different ifindexes
+ */
+LIB_EXPORT bool l_dhcp6_client_set_link_local_address(
+						struct l_dhcp6_client *client,
+						const char *ll)
+{
+	if (unlikely(!client))
+		return false;
+
+	if (inet_pton(AF_INET6, ll, &client->ll_address) != 1)
+		return false;
+
+	return true;
+}
+
 LIB_EXPORT bool l_dhcp6_client_set_debug(struct l_dhcp6_client *client,
 						l_dhcp6_debug_cb_t function,
 						void *user_data,
@@ -1370,24 +1652,26 @@ LIB_EXPORT bool l_dhcp6_client_set_lla_randomized(struct l_dhcp6_client *client,
 	if (unlikely(client->state != DHCP6_STATE_INIT))
 		return false;
 
-	if (client->duid)
-		return false;
-
-	/*
-	 * RFC 7844, Section 4.3 Client Identifier DHCPv6 Option:
-	 * "When using the anonymity profile in conjunction with randomized
-	 * link-layer addresses, DHCPv6 clients MUST use DUID format
-	 * number 3 -- link-layer address.  The value of the link-layer
-	 * address should be the value currently assigned to the interface.
-	 */
-	client->duid_len = 4 + client->addr_len;
-	client->duid = l_malloc(client->duid_len);
-	client->duid->type = L_CPU_TO_BE16(DUID_TYPE_LINK_LAYER_ADDR);
-	l_put_be16(client->addr_type, client->duid->identifier);
-	memcpy(client->duid->identifier + 2, client->addr, client->addr_len);
+	client->lla_randomized = randomized;
 
 	return true;
 }
+
+LIB_EXPORT bool l_dhcp6_client_set_no_rapid_commit(
+						struct l_dhcp6_client *client,
+						bool no_rapid_commit)
+{
+	if (unlikely(!client))
+		return false;
+
+	if (unlikely(client->state != DHCP6_STATE_INIT))
+		return false;
+
+	client->no_rapid_commit = no_rapid_commit;
+
+	return true;
+}
+
 LIB_EXPORT bool l_dhcp6_client_set_nodelay(struct l_dhcp6_client *client,
 						bool nodelay)
 {
@@ -1402,6 +1686,38 @@ LIB_EXPORT bool l_dhcp6_client_set_nodelay(struct l_dhcp6_client *client,
 	return true;
 }
 
+LIB_EXPORT bool l_dhcp6_client_set_nora(struct l_dhcp6_client *client,
+					bool nora)
+{
+	if (unlikely(!client))
+		return false;
+
+	if (unlikely(client->state != DHCP6_STATE_INIT))
+		return false;
+
+	client->nora = nora;
+
+	return true;
+}
+
+/*
+ * Set l_netlink object to use for sending RTNL commands.  If set, then
+ * l_dhcp6_client will add / delete addresses obtained via DHCPv6
+ * automatically.
+ */
+LIB_EXPORT bool l_dhcp6_client_set_rtnl(struct l_dhcp6_client *client,
+						struct l_netlink *rtnl)
+{
+	if (unlikely(!client))
+		return false;
+
+	if (unlikely(client->state != DHCP6_STATE_INIT))
+		return false;
+
+	client->rtnl = rtnl;
+	return true;
+}
+
 LIB_EXPORT bool l_dhcp6_client_set_stateless(struct l_dhcp6_client *client,
 								bool stateless)
 {
@@ -1416,6 +1732,15 @@ LIB_EXPORT bool l_dhcp6_client_set_stateless(struct l_dhcp6_client *client,
 	return true;
 }
 
+LIB_EXPORT struct l_icmp6_client *l_dhcp6_client_get_icmp6(
+						struct l_dhcp6_client *client)
+{
+	if (unlikely(!client))
+		return NULL;
+
+	return client->icmp6;
+}
+
 LIB_EXPORT bool l_dhcp6_client_add_request_option(struct l_dhcp6_client *client,
 						enum l_dhcp6_option option)
 {
@@ -1430,18 +1755,8 @@ LIB_EXPORT bool l_dhcp6_client_add_request_option(struct l_dhcp6_client *client,
 	return true;
 }
 
-static uint64_t pick_delay_interval(uint32_t min_secs, uint32_t max_secs)
-{
-	uint64_t min_ms = min_secs * 1000ULL;
-	uint64_t max_ms = max_secs * 1000ULL;
-
-	return l_getrandom_uint32() % (max_ms + 1 - min_ms) + min_ms;
-}
-
 LIB_EXPORT bool l_dhcp6_client_start(struct l_dhcp6_client *client)
 {
-	uint32_t delay;
-
 	if (unlikely(!client))
 		return false;
 
@@ -1457,11 +1772,15 @@ LIB_EXPORT bool l_dhcp6_client_start(struct l_dhcp6_client *client)
 		l_dhcp6_client_set_address(client, ARPHRD_ETHER, mac, ETH_ALEN);
 	}
 
-	client_duid_generate_addr_plus_time(client);
+	if (client->lla_randomized)
+		client_duid_generate_addr(client);
+	else
+		client_duid_generate_addr_plus_time(client);
 
 	if (!client->transport) {
 		client->transport =
 			_dhcp6_default_transport_new(client->ifindex,
+							&client->ll_address,
 							DHCP6_PORT_CLIENT);
 
 		if (!client->transport)
@@ -1482,40 +1801,26 @@ LIB_EXPORT bool l_dhcp6_client_start(struct l_dhcp6_client *client)
 						dhcp6_client_rx_message,
 						client);
 
-	if (client->stateless) {
-		dhcp6_client_new_transaction(client,
-					DHCP6_STATE_REQUESTING_INFORMATION);
-		delay = pick_delay_interval(0, INF_MAX_DELAY);
-	} else {
-		dhcp6_client_new_transaction(client, DHCP6_STATE_SOLICITING);
-		delay = pick_delay_interval(0, SOL_MAX_DELAY);
+	if (client->nora || client->addr_type != ARPHRD_ETHER) {
+		dhcp6_client_send_initial(client);
+		return true;
 	}
 
-	/*
-	 * RFC 8415, Section 18.2.1:
-	 * "The first Solicit message from the client on the interface SHOULD
-	 * be delayed by a random amount of time between 0 and SOL_MAX_DELAY.
-	 * This random delay helps desynchronize clients that start a DHCP
-	 * session at the same time, such as after recovery from a power
-	 * failure or after a router outage after seeing that DHCP is
-	 * available in Router Advertisement messages..."
-	 *
-	 * Similar verbiage in Section 18.2.6, except it uses 'MUST':
-	 * "The first Information-request message from the client on the
-	 * interface MUST be delayed by a random amount of time between 0 and
-	 * INF_MAX_DELAY."
-	 *
-	 * In certain situations, like when a WiFi network was just joined,
-	 * there's no point in waiting a random time, so we forgo the wait
-	 * if no delay is requested
-	 */
-	if (client->nodelay)
-		dhcp6_client_timeout_send(NULL, client);
-	else
-		client->timeout_send = l_timeout_create_ms(delay,
-						dhcp6_client_timeout_send,
-						client, NULL);
+	l_icmp6_client_set_address(client->icmp6, client->addr);
+	l_icmp6_client_set_debug(client->icmp6, client->debug_handler,
+					client->debug_data,
+					client->debug_destroy);
+	l_icmp6_client_set_nodelay(client->icmp6, client->nodelay);
 
+	if (!l_icmp6_client_start(client->icmp6))
+		return false;
+
+	/*
+	 * Start a timer.  In case there are no router advertisements that
+	 * come back in a reasonable time, fail with a NO_LEASE event
+	 */
+	client->timeout_send = l_timeout_create(10, dhcp6_client_no_ra,
+						client, NULL);
 	return true;
 }
 
@@ -1526,12 +1831,34 @@ LIB_EXPORT bool l_dhcp6_client_stop(struct l_dhcp6_client *client)
 
 	CLIENT_DEBUG("");
 
+	if (client->rtnl_add_cmdid) {
+		l_netlink_cancel(client->rtnl, client->rtnl_add_cmdid);
+		client->rtnl_add_cmdid = 0;
+	}
+
+	if (client->rtnl_configured_address) {
+		l_rtnl_ifaddr_delete(client->rtnl, client->ifindex,
+					client->rtnl_configured_address,
+					NULL, NULL, NULL);
+		l_rtnl_address_free(client->rtnl_configured_address);
+		client->rtnl_configured_address = NULL;
+	}
+
 	_dhcp6_lease_free(client->lease);
 	client->lease = NULL;
 
+	if (!client->nora)
+		l_icmp6_client_stop(client->icmp6);
+
 	l_timeout_remove(client->timeout_send);
 	client->timeout_send = NULL;
 
+	l_timeout_remove(client->timeout_lease);
+	client->timeout_lease = NULL;
+
+	l_free(client->duid);
+	client->duid = NULL;
+
 	if (client->transport && client->transport->close)
 		client->transport->close(client->transport);
 
